from ftw.builder import Builder
from ftw.builder import create
from ftw.keywordwidget.tests import FunctionalTestCase
from ftw.keywordwidget.widget import as_keyword_token
from ftw.keywordwidget.widget import KeywordFieldWidget
from ftw.testbrowser import browsing
from ftw.testbrowser.pages.statusmessages import error_messages
from ftw.testbrowser.pages.statusmessages import info_messages
from zope import schema
import json
import transaction


class TestKeywordWidgetWithRequiredFields(FunctionalTestCase):

    def setUp(self):
        super(TestKeywordWidgetWithRequiredFields, self).setUp()
        self.grant('Manager')
        additional_behaviors = [
            'ftw.keywordwidget.behavior.IKeywordCategorization',
            'ftw.keywordwidget.tests.behavior.IHaveARequiredTextField',
        ]
        self.setup_fti(additional_behaviors=additional_behaviors)

    @browsing
    def test_validation_errors_preserve_user_input(self, browser):
        browser.login().open(view='++add++SampleContent')
        browser.fill(
            {'form.widgets.IKeywordCategorization.subjects_new': 'foo'}
        ).submit()

        self.assertEqual(['There were some errors.'], error_messages())

        new_field = browser.find(
            'form.widgets.IKeywordCategorization.subjects_new')
        self.assertEqual('foo', new_field.value)

        tags = browser.find_field_by_text(u'Tags')
        self.assertEqual(['foo'], tags.options_labels)
        self.assertEqual(['foo'], list(tags.node.value),
                         'expected "foo" to be selected in the widget')

    @browsing
    def test_no_duplicates_after_submitting_after_a_validation_error(self, browser):
        browser.login().open(view='++add++SampleContent')

        browser.fill({
            'form.widgets.IKeywordCategorization.subjects_new': u'N\xe4h'}
        ).submit()
        self.assertEqual(['There were some errors.'], error_messages())

        # the second request will already contain the newly added selected
        # value from the first request (i.e. it's token).
        browser.fill({
            'form.widgets.IKeywordCategorization.subjects': as_keyword_token(
                u'N\xe4h'),
            'form.widgets.IKeywordCategorization.subjects_new': u'N\xe4h',
            'Title': 'jajaja...'}).submit()
        self.assertEqual(['Item created'], info_messages())

        self.assertEqual(['N\xc3\xa4h'], browser.context.subject)


class TestKeywordWidget(FunctionalTestCase):

    def setUp(self):
        super(TestKeywordWidget, self).setUp()
        self.grant('Manager')
        self.setup_fti()

        # Create a content with some keywords
        self.sample = create(Builder('sample content')
                             .titled(u'Dummy content for keywords')
                             .having(subjects=('F\xc3\xb6\xc3\xb6',
                                               'Bar', 'Baz')))

    @browsing
    def test_widget_render_input(self, browser):
        content = create(Builder('sample content').titled(u'A content'))
        browser.login().visit(content, view='edit')

        tags = browser.find_field_by_text(u'Tags')
        self.assertTrue(tags)

        # Possible options
        self.assertListEqual(
            ['Bar', 'Baz', u'F\xf6\xf6'],
            tags.options_labels)

        # New entries field id derives from origin field id.
        new = browser.css('#' + tags.attrib['id'] + '_new')
        self.assertTrue(new)

    @browsing
    def test_select_and_save_from_given_entries(self, browser):
        content = create(Builder('sample content').titled(u'A content'))
        browser.login().visit(content, view='edit')

        form = browser.find_form_by_field('Tags')
        form.fill({'Tags': ('Bar', 'F=C3=B6=C3=B6')}).submit()

        browser.visit(content, view='edit')
        tags = browser.find_field_by_text(u'Tags')
        self.assertEquals(('Bar', 'F=C3=B6=C3=B6'), tuple(tags.value))

    @browsing
    def test_add_new_tokens_to_field(self, browser):
        content = create(Builder('sample content').titled(u'A content'))
        browser.login().visit(content, view='edit')

        tags = browser.find_field_by_text(u'Tags')
        form = browser.find_form_by_field('Tags')
        new = browser.css('#' + tags.attrib['id'] + '_new').first
        new.text = u'NewItem1\nNew Item 2\nN\xf6i 3'
        form.submit()

        browser.visit(content, view='edit')
        tags = browser.find_field_by_text(u'Tags')
        self.assertTupleEqual(('New Item 2', 'NewItem1', 'N=C3=B6i 3'),
                              tuple(tags.value))

    @browsing
    def test_add_new_terms_is_only_available_to_certain_roles(self, browser):
        content = create(Builder('sample content').titled(u'A content'))
        content.manage_permission('ftw.keywordwidget: Add new term',
                                  roles=[],
                                  acquire=0)
        transaction.commit()

        browser.login().visit(content, view='edit')
        tags = browser.find_field_by_text(u'Tags')
        self.assertFalse(len(browser.css('#' + tags.attrib['id'] + '_new')),
                         'Add new term field should NOT be there')

        content.manage_permission('ftw.keywordwidget: Add new term',
                                  roles=['Manager'],
                                  acquire=0)
        transaction.commit()

        browser.visit(content, view='edit')
        tags = browser.find_field_by_text(u'Tags')
        self.assertTrue(browser.css('#' + tags.attrib['id'] + '_new'),
                        'Add new term field should BE there')

    @browsing
    def test_default_select2_config_on_widget(self, browser):
        content = create(Builder('sample content').titled(u'A content'))
        browser.login().visit(content, view='edit')
        tags = browser.find_field_by_text(u'Tags')

        select2_config = json.loads(tags.attrib['data-select2config'])
        self.assertListEqual(
            [u'i18n', u'width', u'allowClear', u'tokenSeparators', u'tags'],
            select2_config.keys())

    @browsing
    def test_add_twice_the_same_term_in_fact_adds_only_one(self, browser):
        content = create(Builder('sample content').titled(u'A content'))
        browser.login().visit(content, view='edit')

        tags = browser.find_field_by_text(u'Tags')
        form = browser.find_form_by_field('Tags')
        new = browser.css('#' + tags.attrib['id'] + '_new').first
        new.text = u'New\nNew'
        form.submit()

        browser.visit(content, view='edit')
        tags = browser.find_field_by_text(u'Tags')
        self.assertTupleEqual(('New', ),
                              tuple(tags.value))

    @browsing
    def test_add_new_term_to_a_give_set_of_tags(self, browser):
        content = create(Builder('sample content')
                         .titled(u'A content')
                         .having(subjects=('Bar', 'Baz')))
        browser.login().visit(content, view='edit')

        tags = browser.find_field_by_text(u'Tags')
        form = browser.find_form_by_field('Tags')
        new = browser.css('#' + tags.attrib['id'] + '_new').first
        new.text = u'New'
        form.submit()

        browser.visit(content, view='edit')
        tags = browser.find_field_by_text(u'Tags')
        self.assertTupleEqual(('Bar', 'Baz', 'New'),
                              tuple(tags.value))

    @browsing
    def test_add_new_value_which_is_already_in_the_source(self, browser):
        content = create(Builder('sample content').titled(u'A content'))
        browser.login().visit(content, view='edit')

        tags = browser.find_field_by_text(u'Tags')
        form = browser.find_form_by_field('Tags')
        new = browser.css('#' + tags.attrib['id'] + '_new').first
        new.text = u'Bar'
        form.submit()

        browser.visit(content, view='edit')
        tags = browser.find_field_by_text(u'Tags')
        self.assertTupleEqual(('Bar', ),
                              tuple(tags.value))

    @browsing
    def test_add_new_term_to_empty_vocabulary(self, browser):
        self.portal.manage_delObjects(ids=[self.sample.getId()])
        transaction.commit()

        content = create(Builder('sample content').titled(u'A content'))
        browser.login().visit(content, view='edit')

        tags = browser.find_field_by_text(u'Tags')
        form = browser.find_form_by_field('Tags')
        new = browser.css('#' + tags.attrib['id'] + '_new').first
        new.text = u'NewItem1\nNew Item 2\nN\xf6i 3'
        form.submit()

        browser.visit(content, view='edit')
        tags = browser.find_field_by_text(u'Tags')
        self.assertTupleEqual(('New Item 2', 'NewItem1', 'N=C3=B6i 3'),
                              tuple(tags.value))

    def test_add_only_one_term_will_not_raise_an_error_after_cleanup_request(self):
        """The widget will not be used trough the browsertest. So we have to do
        it manually with a unittest.
        """
        field = schema.List(title=u"Subjects")
        widget = KeywordFieldWidget(field, self.request)

        widget.name = "Subjects"
        self.request['Subjects_new'] = u"Foo Bar"
        self.request['Subjects'] = u"Foo Bar"

        widget.cleanup_request()

        self.assertEqual([], self.request['Subjects'])
        self.assertEqual('Foo Bar', self.request['Subjects_new'])



from ftw.upgrade import UpgradeStep
import pkg_resources

IS_PLONE_5 = pkg_resources.get_distribution('Products.CMFPlone').version >= '5'


class UpdatePlone5Registry(UpgradeStep):
    """Update Plone5 registry.
    """

    def __call__(self):
        if IS_PLONE_5:
            self.install_upgrade_profile()



from ftw.keywordwidget.field import ChoicePlus
from ftw.keywordwidget.vocabularies import KeywordSearchableAndAddableSourceBinder
from ftw.keywordwidget.vocabularies import KeywordSearchableSourceBinder
from ftw.keywordwidget.widget import KeywordFieldWidget
from plone.app.dexterity.behaviors.metadata import MetadataBase
from plone.autoform import directives
from plone.autoform.interfaces import IFormFieldProvider
from plone.supermodel import model
from Products.CMFPlone import PloneMessageFactory as _PMF
from z3c.form.interfaces import IEditForm, IAddForm
from zope import schema
from zope.interface import alsoProvides


class IKeywordCategorization(model.Schema):

    model.fieldset(
        'categorization',
        label=_PMF(u'label_schema_categorization', default=u'Categorization'),
        fields=['subjects', ],
    )

    directives.widget('subjects', KeywordFieldWidget)
    subjects = schema.List(
        title=_PMF(u'label_tags', default=u'Tags'),
        description=_PMF(
            u'help_tags',
            default=u'Tags are commonly used for ad-hoc organization of ' +
                    u'content.'
        ),
        value_type=ChoicePlus(
            title=u"Multiple",
            vocabulary='plone.app.vocabularies.Keywords', ),
        required=False,
        missing_value=(),
    )

    # Show field only on edit/add forms.
    directives.omitted('subjects', )
    directives.no_omit(IEditForm, 'subjects', )
    directives.no_omit(IAddForm, 'subjects', )

alsoProvides(IKeywordCategorization, IFormFieldProvider)


class Categorization(MetadataBase):

    def _get_subjects(self):
        return self.context.subject

    def _set_subjects(self, value):
        self.context.subject = value
    subjects = property(_get_subjects, _set_subjects)


class IKeywordUseCases(model.Schema):

    directives.widget('types', KeywordFieldWidget)
    types = schema.List(
        title=u'Types',
        value_type=schema.Choice(
            title=u"Multiple",
            vocabulary='plone.app.vocabularies.PortalTypes',
            ),
        required=False,
        missing_value=(),
    )

    directives.widget('types2', KeywordFieldWidget)
    types2 = schema.Choice(
        title=u'Single type',
        vocabulary='plone.app.vocabularies.PortalTypes',
        required=False,
        missing_value=(),
    )

    directives.widget('types3', KeywordFieldWidget)
    types3 = schema.Tuple(
        title=u'Types3',
        value_type=schema.Choice(
            title=u"Multiple",
            vocabulary='plone.app.vocabularies.PortalTypes',
            ),
        required=False,
        missing_value=(),
    )

    directives.widget('unicode_keywords',
                      KeywordFieldWidget,
                      new_terms_as_unicode=True)
    unicode_keywords = schema.Tuple(
        title=u'UnicodeTags',
        value_type=ChoicePlus(
            title=u"Multiple",
            vocabulary='ftw.keywordwidget.UnicodeKeywordVocabulary',
            ),
        required=False,
        missing_value=(),
    )

    directives.widget('async', KeywordFieldWidget, async=True)
    async = schema.Tuple(
        title=u'async',
        value_type=schema.Choice(
            title=u"Multiple",
            source=KeywordSearchableSourceBinder(),
            ),
        required=False,
        missing_value=(),
    )

    directives.widget('async_addable', KeywordFieldWidget, async=True)
    async_addable = schema.Tuple(
        title=u'async_addable',
        value_type=ChoicePlus(
            title=u"Multiple",
            source=KeywordSearchableAndAddableSourceBinder(),
            ),
        required=False,
        missing_value=(),
    )


alsoProvides(IKeywordUseCases, IFormFieldProvider)



from plone import api
from zope.interface import implementer
from zope.schema import Choice
from zope.schema._bootstrapinterfaces import WrongType
from zope.schema.interfaces import ConstraintNotSatisfied
from zope.schema.interfaces import IChoice
from zope.schema.interfaces import IFromUnicode
from zope.schema.vocabulary import getVocabularyRegistry
from zope.schema.vocabulary import VocabularyRegistryError


@implementer(IChoice, IFromUnicode)
class ChoicePlus(Choice):
    """A choice field, which allows additional terms.
    Example: Allow new tags for catalog based vocabulary.
    """

    def _validate(self, value):
        # Pass all validations during initialization
        if self._init_field:
            return

        if self._type is not None and not isinstance(value, self._type):
            raise WrongType(value, self._type, self.__name__)

        if not self.constraint(value):
            raise ConstraintNotSatisfied(value)

        vocabulary = self.vocabulary
        if vocabulary is None:
            vr = getVocabularyRegistry()
            try:
                vocabulary = vr.get(None, self.vocabularyName)
            except VocabularyRegistryError:
                raise ValueError("Can't validate value without vocabulary")

        # The widget can control this attribute too
        if getattr(api.portal.getRequest(), 'allow_new', True):
            # Allow new values!
            return
        else:
            if value not in vocabulary:
                raise ConstraintNotSatisfied(value)



from ftw.builder import Builder
from ftw.builder import create
from ftw.keywordwidget.search import SearchSource
from ftw.keywordwidget.tests import FunctionalTestCase
from ftw.testbrowser import browsing
from zope.interface import implementer
from zope.schema import Choice
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.vocabulary import SimpleVocabulary
import json
import transaction


class TestKeywordWidget(FunctionalTestCase):

    def setUp(self):
        super(TestKeywordWidget, self).setUp()
        self.grant('Manager')
        additional_behaviors = [
            'ftw.keywordwidget.behavior.IKeywordUseCases',
        ]
        self.setup_fti(additional_behaviors=additional_behaviors)
        transaction.commit()

    @browsing
    def test_choose_from_regular_list_with_choice_widget(self, browser):
        content = create(Builder('sample content').titled(u'A content'))
        browser.login().visit(content, view='edit')
        types = browser.find_field_by_text(u'Types')

        self.assertTrue(types)
        # No add new terms field
        self.assertFalse(browser.css('#' + types.attrib['id'] + '_new'))

        # Some Possible options
        self.assertIn('SampleContent',
                      types.options_labels)
        self.assertIn('File',
                      types.options_labels)
        self.assertIn('Link',
                      types.options_labels)

        browser.fill({'Types': ('File', 'SampleContent')}).submit()
        browser.login().visit(content, view='edit')
        types = browser.find_field_by_text(u'Types')
        self.assertEquals(('File', 'SampleContent'), tuple(types.value))

    @browsing
    def test_choose_from_regular_choice_widget(self, browser):
        content = create(Builder('sample content').titled(u'A content'))
        browser.login().visit(content, view='edit')
        types = browser.find_field_by_text(u'Single type')

        self.assertTrue(types)
        # No add new terms field
        self.assertFalse(browser.css('#' + types.attrib['id'] + '_new'))

        # Some Possible options
        self.assertIn('SampleContent',
                      types.options_labels)
        self.assertIn('File',
                      types.options_labels)
        self.assertIn('Link',
                      types.options_labels)

        browser.fill({'Single type': 'SampleContent'}).submit()
        browser.login().visit(content, view='edit')
        types = browser.find_field_by_text(u'Single type')
        self.assertEquals('SampleContent', types.value)

    @browsing
    def test_choose_from_tuple_with_choice_widget(self, browser):
        content = create(Builder('sample content').titled(u'A content'))
        browser.login().visit(content, view='edit')
        types = browser.find_field_by_text(u'Types3')

        self.assertTrue(types)
        # No add new terms field
        self.assertFalse(browser.css('#' + types.attrib['id'] + '_new'))

        # Some Possible options
        self.assertIn('SampleContent',
                      types.options_labels)
        self.assertIn('File',
                      types.options_labels)
        self.assertIn('Link',
                      types.options_labels)

        browser.fill({'Types3': ('File', 'SampleContent')}).submit()
        browser.login().visit(content, view='edit')
        types = browser.find_field_by_text(u'Types3')
        self.assertEquals(('File', 'SampleContent'), tuple(types.value))

    @browsing
    def test_adding_terms_with_unicode_values_and_vocab_with_unicode_values(self, browser):
        content = create(Builder('sample content').titled(u'A content'))
        browser.login().visit(content, view='edit')

        tags = browser.find_field_by_text(u'UnicodeTags')
        form = browser.find_form_by_field('UnicodeTags')
        new = browser.css('#' + tags.attrib['id'] + '_new').first
        new.text = u'NewItem1\nNew Item 2\nN\xf6i 3'
        form.submit()

        browser.visit(content, view='edit')
        tags = browser.find_field_by_text(u'UnicodeTags')
        self.assertTupleEqual(('New Item 2', 'NewItem1', 'N=C3=B6i 3'),
                              tuple(tags.value))


class TestAsyncOption(FunctionalTestCase):
    def setUp(self):
        super(TestAsyncOption, self).setUp()
        self.grant('Manager')
        additional_behaviors = [
            'ftw.keywordwidget.behavior.IKeywordCategorization',
            'ftw.keywordwidget.behavior.IKeywordUseCases',
        ]
        self.setup_fti(additional_behaviors=additional_behaviors)
        transaction.commit()

    @browsing
    def test_async_render_only_selected_items(self, browser):
        create(Builder('sample content')
               .titled(u'A content')
               .having(subjects=('foo', 'bar', 'baz', 'abc',
                                 'zzzzzz', 'lorem', 'ipsum')))

        content = create(Builder('sample content')
                         .titled(u'Use async lib')
                         .having(async=('abc', 'zzzzzz', 'lorem')))

        browser.login().visit(content, view='edit')
        field = browser.find_field_by_text('async')
        self.assertListEqual(list(field.value), field.options_values)

    @browsing
    def test_async_option_is_selectable_but_not_rendered(self, browser):
        create(Builder('sample content')
               .titled(u'A content')
               .having(subjects=('foo', 'bar', 'baz', 'abc',
                                 'zzzzzz', 'lorem', 'ipsum')))

        content = create(Builder('sample content')
                         .titled(u'Use async lib'))

        browser.login().visit(content, view='edit')
        field = browser.find_field_by_text('async')
        self.assertFalse(tuple(field.value), 'Expect no selectable options')

        form = browser.find_form_by_field('async')
        form.find_widget('async').fill(['foo', 'bar'])
        form.submit()

        self.assertSequenceEqual(['foo', 'bar'], content.async)

    @browsing
    def test_async_option_terms_not_in_source_are_ignored(self, browser):
        create(Builder('sample content')
               .titled(u'A content')
               .having(subjects=('foo', 'bar', 'baz', 'abc',
                                 'zzzzzz', 'lorem', 'ipsum')))

        content = create(Builder('sample content')
                         .titled(u'Use async lib'))

        browser.login().visit(content, view='edit')

        form = browser.find_form_by_field('async')
        form.find_widget('async').fill(['New', 'New2'])
        form.submit()

        self.assertFalse(content.async, 'Expect nothing in field async')

    def test_async_option_only_works_with_IQuerySource(self):

        @implementer(IContextSourceBinder)
        class DummySource(object):
            def __call__(self, context):
                return SimpleVocabulary([])

        content = create(Builder('sample content')
                         .titled(u'A content')
                         .having(subjects=('foo', 'bar', 'baz', 'abc')))

        edit_form = content.restrictedTraverse('@@edit').form_instance
        edit_form.update()
        new_field = Choice(title=u'dummy', source=DummySource())
        edit_form.fields['IKeywordUseCases.async'].field = new_field

        with self.assertRaises(TypeError):
            edit_form.update()

    def test_search_endpoint(self):
        content = create(Builder('sample content')
                         .titled(u'A content')
                         .having(subjects=('foo', 'bar', 'baz', 'abc',
                                           'zzzzzz', 'lorem', 'ipsum')))

        self.portal.REQUEST.set('q', 'ba')
        result = json.loads(self._get_search_view(content))

        self.assertTrue(isinstance(result, dict))
        self.assertIn('total_count', result)
        self.assertEquals(result['total_count'], 2)

        self.assertIn('page', result)
        self.assertEquals(result['page'], 1)

        self.assertIn('results', result)
        self.assertEquals([{'_resultId': u'bar',
                            'id': u'bar',
                            'text': u'bar'},
                           {'_resultId': u'baz',
                            'id': u'baz',
                            'text': u'baz'}, ],
                          result['results'])

        self.portal.REQUEST.set('q', 'dummy')
        result = json.loads(self._get_search_view(content))
        self.assertEquals([], result['results'])

    def test_search_endpoint_batch_result(self):
        content = create(Builder('sample content')
                         .titled(u'A content')
                         .having(subjects=map(str, range(1, 100))))

        self.portal.REQUEST.set('q', '1')
        self.portal.REQUEST.set('pagesize', '5')
        self.portal.REQUEST.set('page', '1')
        result = json.loads(self._get_search_view(content))

        self.assertEquals(5, len(result['results']))
        self.assertEquals(19, result['total_count'])
        self.assertEquals([u'1', u'10', u'11', u'12', u'13'],
                          self._get_values(result['results']))

        self.portal.REQUEST.set('q', '1')
        self.portal.REQUEST.set('pagesize', '5')
        self.portal.REQUEST.set('page', '3')
        result = json.loads(self._get_search_view(content))

        self.assertEquals(5, len(result['results']))
        self.assertEquals(19, result['total_count'])
        self.assertEquals([u'19', u'21', u'31', u'41', u'51'],
                          self._get_values(result['results']))

    def _get_values(self, items):
        return map(lambda item: item['id'], items)

    def _get_search_view(self, context):
        form = context.unrestrictedTraverse('@@edit').form_instance
        form.update()
        widget = form.widgets['IKeywordUseCases.async']
        return SearchSource(widget, widget.request)()


class TestAsyncAddableOption(FunctionalTestCase):

    def setUp(self):
        super(TestAsyncAddableOption, self).setUp()
        self.grant('Manager')
        additional_behaviors = [
            'ftw.keywordwidget.behavior.IKeywordCategorization',
            'ftw.keywordwidget.behavior.IKeywordUseCases',
        ]
        self.setup_fti(additional_behaviors=additional_behaviors)
        transaction.commit()

    @browsing
    def test_async_render_only_selected_items(self, browser):
        create(Builder('sample content')
               .titled(u'A content')
               .having(subjects=('foo', 'bar', 'baz', 'abc',
                                 'zzzzzz', 'lorem', 'ipsum')))

        content = create(Builder('sample content')
                         .titled(u'Use async lib')
                         .having(async_addable=('abc', 'zzzzzz', 'lorem')))

        browser.login().visit(content, view='edit')
        field = browser.find_field_by_text('async_addable')
        self.assertListEqual(list(field.value), field.options_values)

    @browsing
    def test_async_option_is_selectable_but_not_rendered(self, browser):
        create(Builder('sample content')
               .titled(u'A content')
               .having(subjects=('foo', 'bar', 'baz', 'abc',
                                 'zzzzzz', 'lorem', 'ipsum')))

        content = create(Builder('sample content')
                         .titled(u'Use async lib'))

        browser.login().visit(content, view='edit')
        field = browser.find_field_by_text('async_addable')
        self.assertFalse(tuple(field.value), 'Expect no selectable options')

        form = browser.find_form_by_field('async_addable')
        form.find_widget('async_addable').fill(['foo', 'bar'])
        form.submit()

        self.assertSequenceEqual(['foo', 'bar'], content.async_addable)

    @browsing
    def test_async_option_terms_not_in_source_are_ignored(self, browser):
        create(Builder('sample content')
               .titled(u'A content')
               .having(subjects=('foo', 'bar', 'baz', 'abc',
                                 'zzzzzz', 'lorem', 'ipsum')))

        content = create(Builder('sample content')
                         .titled(u'Use async lib'))

        browser.login().visit(content, view='edit')

        form = browser.find_form_by_field('async_addable')
        form.find_widget('async_addable').fill(['New', 'New2'])
        form.submit()

        self.assertFalse(content.async_addable, 'Expect nothing in field async')

    @browsing
    def test_can_add_terms_in_async_mode(self, browser):
        content = create(Builder('sample content').titled(u'A content'))
        browser.login().visit(content, view='edit')

        tags = browser.find_field_by_text(u'async_addable')
        form = browser.find_form_by_field('async_addable')
        new = browser.css('#' + tags.attrib['id'] + '_new').first
        new.text = u'NewItem1\nNew Item 2\nN\xf6i 3'
        form.submit()

        self.assertItemsEqual(('New Item 2', 'NewItem1', 'N\xc3\xb6i 3'),
                              content.async_addable)

    def test_async_option_only_works_with_IQuerySource(self):

        @implementer(IContextSourceBinder)
        class DummySource(object):
            def __call__(self, context):
                return SimpleVocabulary([])

        content = create(Builder('sample content')
                         .titled(u'A content')
                         .having(subjects=('foo', 'bar', 'baz', 'abc')))

        edit_form = content.restrictedTraverse('@@edit').form_instance
        edit_form.update()
        new_field = Choice(title=u'dummy', source=DummySource())
        edit_form.fields['IKeywordUseCases.async_addable'].field = new_field

        with self.assertRaises(TypeError):
            edit_form.update()

    def test_search_endpoint(self):
        content = create(Builder('sample content')
                         .titled(u'A content')
                         .having(subjects=('foo', 'bar', 'baz', 'abc',
                                           'zzzzzz', 'lorem', 'ipsum')))

        self.portal.REQUEST.set('q', 'ba')
        result = json.loads(self._get_search_view(content))

        self.assertTrue(isinstance(result, dict))
        self.assertIn('total_count', result)
        self.assertEquals(result['total_count'], 2)

        self.assertIn('page', result)
        self.assertEquals(result['page'], 1)

        self.assertIn('results', result)
        self.assertEquals([{'_resultId': u'bar',
                            'id': u'bar',
                            'text': u'bar'},
                           {'_resultId': u'baz',
                            'id': u'baz',
                            'text': u'baz'}, ],
                          result['results'])

        self.portal.REQUEST.set('q', 'dummy')
        result = json.loads(self._get_search_view(content))
        self.assertEquals([], result['results'])

    def test_search_endpoint_batch_result(self):
        content = create(Builder('sample content')
                         .titled(u'A content')
                         .having(subjects=map(str, range(1, 100))))

        self.portal.REQUEST.set('q', '1')
        self.portal.REQUEST.set('pagesize', '5')
        self.portal.REQUEST.set('page', '1')
        result = json.loads(self._get_search_view(content))

        self.assertEquals(5, len(result['results']))
        self.assertEquals(19, result['total_count'])
        self.assertEquals([u'1', u'10', u'11', u'12', u'13'],
                          self._get_values(result['results']))

        self.portal.REQUEST.set('q', '1')
        self.portal.REQUEST.set('pagesize', '5')
        self.portal.REQUEST.set('page', '3')
        result = json.loads(self._get_search_view(content))

        self.assertEquals(5, len(result['results']))
        self.assertEquals(19, result['total_count'])
        self.assertEquals([u'19', u'21', u'31', u'41', u'51'],
                          self._get_values(result['results']))

    def _get_values(self, items):
        return map(lambda item: item['id'], items)

    def _get_search_view(self, context):
        form = context.unrestrictedTraverse('@@edit').form_instance
        form.update()
        widget = form.widgets['IKeywordUseCases.async_addable']
        return SearchSource(widget, widget.request)()



from ftw.builder import registry
from ftw.builder.dexterity import DexterityBuilder
from ftw.keywordwidget.testing import FTW_FUNCTIONAL
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.dexterity.content import Container
from plone.dexterity.fti import DexterityFTI
from plone.supermodel import model
from unittest import TestCase
from zope.interface import implements
from zope.interface import Interface
import ftw.keywordwidget.tests.widget
import transaction


class ISampleContentSchema(model.Schema):
    pass


class ISampleContententMarker(Interface):
    pass


class SampleContent(Container):
    implements(ISampleContententMarker)


class FunctionalTestCase(TestCase):
    layer = FTW_FUNCTIONAL

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']

    def grant(self, *roles):
        setRoles(self.portal, TEST_USER_ID, list(roles))
        transaction.commit()

    def setup_fti(self, additional_behaviors=None):
        types_tool = self.portal.portal_types

        default_behaviors = [
            'plone.app.dexterity.behaviors.metadata.IBasic',
            'plone.app.content.interfaces.INameFromTitle'
        ]

        if additional_behaviors is None:
            default_behaviors += [
                'ftw.keywordwidget.behavior.IKeywordCategorization',
            ]
        else:
            default_behaviors += additional_behaviors

        fti = DexterityFTI('SampleContent')
        fti.schema = 'ftw.keywordwidget.tests.ISampleContentSchema'
        fti.klass = 'ftw.keywordwidget.tests.SampleContent'
        fti.behaviors = tuple(default_behaviors)
        fti.default_view = 'view'
        types_tool._setObject('SampleContent', fti)

        transaction.commit()


# Builders
class SampleContentBuilder(DexterityBuilder):
    portal_type = 'SampleContent'

registry.builder_registry.register(
    'sample content', SampleContentBuilder)



from plone.batching.batch import Batch
from Products.Five.browser import BrowserView
import json


class SearchSource(BrowserView):

    def __init__(self, context, request):
        super(SearchSource, self).__init__(context, request)
        self.widget = self.context

    def __call__(self):
        self.request.response.setHeader('Content-Type', 'application/json')
        self.request.response.setHeader('X-Theme-Disabled', 'True')

        query = self.request.get('q', None)
        page = int(self.request.get('page', 1))
        pagesize = int(self.request.get('pagesize', 20))

        if not query:
            return json.dumps({})

        self.widget.update()
        source = self.widget.choice_field.source(self.widget.context)

        batch = Batch.fromPagenumber(items=source.search(query),
                                     pagesize=pagesize,
                                     pagenumber=page)

        return json.dumps(
            {
                'results': map(self._term_to_dict, batch),
                'total_count': len(batch),
                'page': page,
                'pagination': {'more': (page * pagesize) < len(batch)}
            }
        )

    def _term_to_dict(self, term):
        return {'_resultId': term.token,
                'id': term.token,
                'text': term.title and term.title or term.token}



from ftw.testbrowser.widgets.base import PloneWidget
from ftw.testbrowser.widgets.base import widget
from lxml import etree


@widget
class AsyncKeywordWidget(PloneWidget):

    @staticmethod
    def match(node):
        if not node.tag == 'div' or 'field' not in node.classes:
            return False
        return bool(node.css('.keyword-widget[data-ajaxoptions*="url"]'))

    def fill(self, values):
        select = self.css('select').first
        multiple = select.attrib.get('multiple', None) == 'multiple'

        if isinstance(values, basestring):
            values = [values]

        if not multiple and len(values) > 1:
            raise ValueError('Expect only one item, since the field is '
                             'single valued')

        # Clear options
        for item in self.css('option'):
            item.node.getparent().remove(item.node)

        for value in values:
            option = etree.Element('option')
            option.attrib['value'] = value
            option.attrib['selected'] = 'selected'
            option.text = value
            select.append(option)



from ftw.upgrade import UpgradeStep


class DoNoCompressTheAlreadyMinifiedSelect2MinJs(UpgradeStep):
    """Disable compression of the already minified "select2.min.js".
    """

    def __call__(self):
        self.install_upgrade_profile()



from ftw.builder import Builder
from ftw.builder import create
from ftw.keywordwidget.tests import FunctionalTestCase
from ftw.keywordwidget.vocabularies import KeywordSearchableSource
from ftw.keywordwidget.vocabularies import UnicodeKeywordsVocabulary
import transaction


class TestKeywordWidget(FunctionalTestCase):

    def setUp(self):
        super(TestKeywordWidget, self).setUp()
        self.grant('Manager')
        self.setup_fti()
        transaction.commit()

    def test_unicode_keyword_vocab_with_existing_index(self):
        create(Builder('sample content')
               .titled(u'A content')
               .having(subjects=(u'Bar', u'Baz')))

        vocab = UnicodeKeywordsVocabulary(index_name='Subject')(self.portal)
        self.assertItemsEqual([u'Bar', u'Baz'], vocab.by_value.keys())

    def test_unicode_keyword_vocab_with_NOT_existing_index(self):
        create(Builder('sample content')
               .titled(u'A content')
               .having(subjects=('Bar', 'Baz')))

        vocab = UnicodeKeywordsVocabulary(index_name='dummy')(self.portal)
        self.assertFalse(vocab.by_value.keys(), 'Expect an empty vocab')

    def test_keyword_searchable_source_terms(self):
        create(Builder('sample content')
               .titled(u'A content')
               .having(subjects=(u'Bar', u'Baz')))

        vocab = KeywordSearchableSource(self.portal)
        term = vocab.getTermByToken('Bar')
        self.assertEqual('Bar', term.title)
        self.assertEqual('Bar', term.value)
        self.assertEqual('Bar', term.token)



from ftw.keywordwidget import _
from ftw.keywordwidget.field import ChoicePlus
from ftw.keywordwidget.utils import as_keyword_token
from ftw.keywordwidget.utils import safe_utf8
from ftw.keywordwidget.vocabularies import IKeywordWidgetAddableSource
from plone import api
from Products.CMFPlone.utils import safe_unicode
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from z3c.form.browser.select import SelectWidget
from z3c.form.interfaces import DISPLAY_MODE
from z3c.form.interfaces import HIDDEN_MODE
from z3c.form.interfaces import IFieldWidget
from z3c.form.interfaces import INPUT_MODE
from z3c.form.interfaces import ISelectWidget
from z3c.form.interfaces import NOVALUE
from z3c.form.widget import FieldWidget
from z3c.formwidget.query.interfaces import IQuerySource
from zope import schema
from zope.i18n import translate
from zope.interface import implementer
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.interfaces import ITitledTokenizedTerm
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary
import json


def is_list_type_field(field):
    if isinstance(field, schema.List):
        return True
    elif isinstance(field, schema.Tuple):
        return True

    return False


class IKeywordWidget(ISelectWidget):
    """Marker interface for the Keywordwidget"""


@implementer(IKeywordWidget)
class KeywordWidget(SelectWidget):

    klass = u'keyword-widget'

    noValueToken = u''

    noValueMessage = _('no value')
    promptMessage = _('select some values ...')
    promptNoresultFound = _('No result found')
    label_new = _(u'New')
    label_searching = _(u'Searching...')
    label_loading_more = _('Load more results...')
    label_tooshort_prefix = _('Please enter ')
    label_tooshort_postfix = _(' or more characters')

    multiple = 'multiple'
    size = 10

    # KeywordWidget specific
    js_config = None
    config_json = ''
    ajax_options_json = ''
    choice_field = None
    new_terms_as_unicode = None
    async = False
    template_selection = ''
    template_result = ''

    display_template = ViewPageTemplateFile('templates/keyword_display.pt')
    input_template = ViewPageTemplateFile('templates/keyword_input.pt')
    hidden_template = ViewPageTemplateFile('templates/keyword_hidden.pt')

    def __init__(self, request, js_config=None, new_terms_as_unicode=False,
                 async=False, template_selection='', template_result=''):
        self.request = request
        self.js_config = js_config
        self.new_terms_as_unicode = new_terms_as_unicode
        self.async = async
        self.template_result = template_result
        self.template_selection = template_selection

        # Do not make this dynamic or you will have to handle flip-flops
        # between multiple widgets - we cache the security on the request
        self.add_permission = 'ftw.keywordwidget: Add new term'

    def render(self):
        if self.mode == INPUT_MODE:
            return self.input_template(self)
        elif self.mode == DISPLAY_MODE:
            return self.display_template(self)
        elif self.mode == HIDDEN_MODE:
            return self.hidden_template(self)
        raise NotImplementedError(
            'Mode: "{0}" not supported'.format(self.mode))

    def get_new_values_from_request(self):
        """Get the values from the request and splits addition values by
        newlines,
        """
        new_values = set()
        new = self.request.get('{0}_new'.format(self.name), '').split('\n')
        for value in new:
            cleanedup_value = value.strip().strip('\r').strip('\n')
            if cleanedup_value:
                new_values.add(cleanedup_value)
        return list(new_values)

    def cleanup_request(self):
        """
        If the tags feature is enabled the new term is also in a
        "not normalized" variation in the request. We need to remove it, thus
        the super call of the sequence widget will still work.
        Otherwise it does not recognize all terms in the request as valid term.
        """
        values = self.request.get(self.name, [])
        if not isinstance(values, list):
            # This case happens if the user tries to add only one new keyword.
            # We will receive a string instead a list.
            values = [values]

        if not values:
            return

        for new_value in self.get_new_values_from_request():
            if new_value in values:
                values.remove(new_value)
        self.request.set(self.name, values)

    def _validate_source_for_async(self):
        source = self.choice_field.source
        if IContextSourceBinder.providedBy(source):
            source = source(self.context)
            if IQuerySource.providedBy(source):
                return
        raise TypeError('A IContextSourceBinder with IQuerySource is needed')

    def update(self):
        self.get_choice_field()

        if self.async:
            self._validate_source_for_async()

        super(KeywordWidget, self).update()

        self.update_multivalued_property()

        if isinstance(self.choice_field, ChoicePlus):
            api.portal.getRequest().allow_new = api.user.has_permission(
                self.add_permission,
                obj=self.context,
                )

        self.update_js_config()

    def update_js_config(self):
        # Sane default config
        default_config = {
            'i18n': {
                'label_placeholder': translate(
                    self.promptMessage,
                    context=self.request,
                    ),
                'label_no_result': translate(
                    self.promptNoresultFound,
                    context=self.request,
                    ),
                'label_new': translate(
                    self.label_new,
                    context=self.request,
                    ),
                'label_searching': translate(
                    self.label_searching,
                    context=self.request,
                    ),
                'label_loading_more': translate(
                    self.label_loading_more,
                    context=self.request,
                    ),
                'label_tooshort_prefix': translate(
                    self.label_tooshort_prefix,
                    context=self.request,
                    ),
                'label_tooshort_postfix': translate(
                    self.label_tooshort_postfix,
                    context=self.request,
                    ),
            },
            'width': '300px',
            'allowClear': not self.field.required and not self.multiple,
        }

        if self.show_add_term_field():
            default_config['tags'] = True
            default_config['tokenSeparators'] = [',']

        if self.async:
            default_config['minimumInputLength'] = 3

        if self.js_config:
            default_config.update(self.js_config)

        if self.async:
            self.ajax_options_json = json.dumps(
                {'url': '{}/++widget++{}/search'.format(self.request.getURL(),
                                                        self.name),
                 'dataType': 'json',
                 'delay': 250}
            )

        self.config_json = json.dumps(default_config)

    def update_multivalued_property(self):
        if not is_list_type_field(self.field):
            self.multiple = None
            self.size = 1
            self.promptMessage = _('select a value ...')

    def get_choice_field(self):
        is_list = is_list_type_field(self.field)
        is_choice = isinstance(self.field, schema.Choice)

        if is_list and isinstance(self.field.value_type, schema.Choice):
            self.choice_field = self.field.value_type
        elif is_choice:
            self.choice_field = self.field
        else:
            raise ValueError('The field or the value type of the field '
                             'needs to be a Choice field.')

    def show_add_term_field(self):
        if isinstance(self.choice_field, ChoicePlus):
            return api.user.has_permission(
                self.add_permission,
                obj=self.context,
                )

        return False

    def extract(self, default=NOVALUE):
        """See z3c.form.interfaces.IWidget.
        """
        self.cleanup_request()

        values = super(KeywordWidget, self).extract(default=default)
        if values is NOVALUE or not self.show_add_term_field():
            return values

        # Adding new keywords, which are not in the vocab/source
        if (self.name + '_new' not in self.request and
                self.name + '-empty-marker' in self.request):
            return []

        tokens = set(values)
        for new_value in self.get_new_values_from_request():
            tokens.add(as_keyword_token(new_value))

        return list(tokens) if tokens else default

    def updateTerms(self):
        super(KeywordWidget, self).updateTerms()

        vocabulary = self.terms.terms

        if self.async and not IKeywordWidgetAddableSource.providedBy(vocabulary):
            # It's only possible to add new terms in async mode if the source
            # implements IKeywordWidgetAddableSource
            return self.terms
        elif IKeywordWidgetAddableSource.providedBy(vocabulary):
            simple_vocabulary = self.terms.terms.instance_vocabulary
        else:
            simple_vocabulary = self.terms.terms

        terms = simple_vocabulary._terms
        for new_value in self.get_new_values_from_request():
            if new_value not in simple_vocabulary:
                # Vocabulary term tokens *must* be 7 bit values, titles *must*
                # be unicode.
                # Value needs to be a utf-8 str, only hell knows why.
                # IMHO this should depend on the source. We're gonna have
                # trouble with this in the future.
                # ...
                # Well as I said before...
                new_token = as_keyword_token(new_value)
                if self.new_terms_as_unicode:
                    new_value = safe_unicode(new_value)
                else:
                    new_value = safe_utf8(new_value)

                terms.append(
                    SimpleTerm(new_value, new_token, safe_unicode(new_value)))

        if IKeywordWidgetAddableSource.providedBy(vocabulary):
            self.terms.terms.instance_vocabulary = SimpleVocabulary(terms)
        else:
            self.terms.terms = SimpleVocabulary(terms)

        return self.terms

    def items(self):
        """This is a copy of the z3c.form.browser.select.SelectWidget
        There's one difference, if the widget is in async mode.
        The widget actually only renders the select values and not all
        possible values.
        """
        if self.terms is None:  # update() has not been called yet
            return ()
        items = []
        if (not self.required or self.prompt) and self.multiple is None:
            message = self.noValueMessage
            if self.prompt:
                message = self.promptMessage
            items.append({
                'id': self.id + '-novalue',
                'value': self.noValueToken,
                'content': message,
                'selected': self.value == []
            })

        ignored = set(self.value)

        def addItem(idx, term, prefix=''):
            selected = self.isSelected(term)
            if selected and term.token in ignored:
                ignored.remove(term.token)
            item_id = '%s-%s%i' % (self.id, prefix, idx)
            content = term.token
            if ITitledTokenizedTerm.providedBy(term):
                content = translate(
                    term.title, context=self.request, default=term.title)
            items.append(
                {'id': item_id, 'value': term.token, 'content': content,
                 'selected': selected})

        if self.async:
            terms = [self.terms.getTermByToken(v) for v in self.value if v]
        else:
            terms = self.terms

        for idx, term in enumerate(terms):
            addItem(idx, term)

        if ignored:
            # some values are not displayed, probably they went away from the
            # vocabulary
            for idx, token in enumerate(sorted(ignored)):
                try:
                    term = self.terms.getTermByToken(token)
                except LookupError:
                    # just in case the term really went away
                    continue

                addItem(idx, term, prefix='missing-')
        return items


@implementer(IFieldWidget)
def KeywordFieldWidget(field, request):
    return FieldWidget(field, KeywordWidget(request))



from Products.CMFCore.utils import getToolByName
from ftw.keywordwidget.tests import FunctionalTestCase
from ftw.testing import IS_PLONE_5
from plone import api


class TestDefaultProfile(FunctionalTestCase):

    def test_installed(self):
        portal_setup = getToolByName(self.portal, 'portal_setup')
        version = portal_setup.getLastVersionForProfile('ftw.keywordwidget:default')
        self.assertNotEqual(version, None)
        if IS_PLONE_5:
            resources = api.portal.get_registry_record('plone.bundles/ftw-keywordwidget-resources.jscompilation')
            self.assertTrue(bool(resources), "Profile not installed.")
        else:
            self.assertNotEqual(version, 'unknown')



from ftw.builder.testing import BUILDER_LAYER
from ftw.builder.testing import functional_session_factory
from ftw.builder.testing import set_builder_session_factory
from ftw.keywordwidget.behavior import IKeywordUseCases
from ftw.testing import IS_PLONE_5
from ftw.testing.layer import COMPONENT_REGISTRY_ISOLATION
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import PloneSandboxLayer
from plone.indexer import indexer
from plone.testing import z2
from Products.CMFCore.interfaces import IContentish
from zope.component import provideAdapter
from zope.configuration import xmlconfig


def _setup_catalog_for_tests(portal):
    @indexer(IContentish)
    def unicode_keywords(obj):
        return IKeywordUseCases(obj).unicode_keywords

    provideAdapter(unicode_keywords)

    catalog = portal.portal_catalog
    catalog.addIndex('unicode_keywords', 'KeywordIndex')


class FtwLayer(PloneSandboxLayer):
    defaultBases = (COMPONENT_REGISTRY_ISOLATION, BUILDER_LAYER)

    def setUpZope(self, app, configurationContext):
        xmlconfig.string(
            '<configure xmlns="http://namespaces.zope.org/zope">'
            '  <include package="z3c.autoinclude" file="meta.zcml" />'
            '  <includePlugins package="plone" />'
            '  <includePluginsOverrides package="plone" />'
            '</configure>',
            context=configurationContext)

        z2.installProduct(app, 'ftw.keywordwidget')

        import ftw.keywordwidget.tests
        xmlconfig.file('tests.zcml',
                       ftw.keywordwidget.tests,
                       context=configurationContext)

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'ftw.keywordwidget:default')
        _setup_catalog_for_tests(portal)

        if IS_PLONE_5:
            applyProfile(portal, 'plone.app.contenttypes:default')


FTW_FIXTURE = FtwLayer()
FTW_FUNCTIONAL = FunctionalTesting(
    bases=(FTW_FIXTURE,
           set_builder_session_factory(functional_session_factory)),
    name="ftw.keywordwidget:functional")



from binascii import b2a_qp


def safe_utf8(text):
    if isinstance(text, unicode):
        text = text.encode('utf8')
    return text


def as_keyword_token(value):
    value = safe_utf8(value)
    return b2a_qp(value)



from setuptools import setup, find_packages
import os

version = '2.2.4'
maintainer = '4teamwork'

tests_require = [
    'ftw.builder',
    'ftw.testing',
    'ftw.testbrowser',
    'plone.app.contenttypes',
    'plone.app.testing',
    'plone.testing',
]

extras_require = {
    'tests': tests_require,
}


setup(
    name='ftw.keywordwidget',
    version=version,
    description='A z3c form keyword widget using select2.',
    long_description=open('README.rst').read() + '\n' +
        open(os.path.join('docs', 'HISTORY.txt')).read(),

    # Get more strings from
    # http://www.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Framework :: Plone',
        'Framework :: Plone :: 4.3',
        'Framework :: Plone :: 5.1',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    keywords='ftw keyword widget',
    author='4teamwork AG',
    author_email='mailto:info@4teamwork.ch',
    maintainer=maintainer,
    url='https://github.com/4teamwork/ftw.keywordwidget',
    license='GPL2',

    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=['ftw'],
    include_package_data=True,
    zip_safe=False,

    install_requires=[
        'ftw.upgrade',
        'setuptools',
        'plone.dexterity',
        'plone.app.dexterity',
        'plone.app.vocabularies',
        'plone.api',
        'Plone',
    ],

    tests_require=tests_require,
    extras_require=extras_require,

    entry_points="""
    # -*- Entry points: -*-
    [z3c.autoinclude.plugin]
    target = plone
    """,
)



import pkg_resources


IS_PLONE_5 = pkg_resources.get_distribution('Products.CMFPlone').version >= '5'


if IS_PLONE_5:
    from Products.CMFCore.indexing import processQueue
else:
    # optional collective.indexing support
    try:
        from collective.indexing.queue import processQueue
    except ImportError:
        def processQueue():
            pass



from binascii import b2a_qp
from ftw.keywordwidget.compat import processQueue
from ftw.keywordwidget.utils import as_keyword_token
from ftw.keywordwidget.utils import safe_utf8
from itertools import chain
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import safe_unicode
from z3c.formwidget.query.interfaces import IQuerySource
from zope.component.hooks import getSite
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.interface import Interface
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.interfaces import IVocabularyFactory
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


@implementer(IVocabularyFactory)
class UnicodeKeywordsVocabulary(object):
    """ This is a copy of plone.app.vocabularies.catalog.KeywordsVocabulary
    with one significant difference the values of the vocabulary terms are
    unicode instead of utf-8. The catalog should return utf-8 by convention.

    Plus it allows you to choose the index_name, which is used to build the
    vocabulary.
    """
    index_name = 'unicode_keywords'

    def __init__(self, index_name=None):
        if index_name:
            self.index_name = index_name

    def __call__(self, context):
        site = getSite()

        # XXX: Explicitly flush collective.indexing queue to make sure we
        # work with up-to-date results.
        processQueue()

        self.catalog = getToolByName(site, "portal_catalog", None)
        if self.catalog is None:
            return SimpleVocabulary([])

        if self.index_name in self.catalog._catalog.indexes:
            index = self.catalog._catalog.getIndex(self.index_name)
            # Vocabulary term tokens *must* be 7 bit values, titles *must* be
            # unicode
            items = [
                SimpleTerm(safe_unicode(i),
                           b2a_qp(safe_utf8(i)),
                           safe_unicode(i))
                for i in index._index
            ]
            return SimpleVocabulary(items)

        else:
            return SimpleVocabulary([])

UnicodeKeywordsVocabularyFactory = UnicodeKeywordsVocabulary()


class IKeywordWidgetAddableSource(Interface):

    def getTermByToken(token):
        """Checks whether the term exists in the current instance
        of the vocabulary. This allows new terms that have been added
        through addTermToInstance to pass validation.
        """


class KeywordWidgetAddableSourceWrapper(object):
    """Wrapper class used to allow adding new terms in async mode
    to a source vocabulary
    """

    def __init__(self, source):
        # Adding an interface to source is not ideal, as this means that
        # wrapping actually modifies source. This is not a problem though
        # as the wrapper is called in the binder, when the vocabulary is
        # instantiated.
        alsoProvides(source, IKeywordWidgetAddableSource)
        self._source = source
        self.instance_vocabulary = SimpleVocabulary([])

    def getTermByToken(self, token):
        """Checks whether the term exists in the current instance_vocabulary.
        This allows new terms that have been added in this request to pass
        validation.
        """
        try:
            return self._source.getTermByToken(token)
        except LookupError:
            return self.instance_vocabulary.getTermByToken(token)

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self._source, attr)

    def __iter__(self):
        return chain(self._source, self.instance_vocabulary)

    def __contains__(self, value):
        return value in chain(self._source, self.instance_vocabulary)


@implementer(IQuerySource)
class KeywordSearchableSource(object):
    """This example of a IQuerySource is taken from the
    plone.formwidget.autocomplete
    """

    def __init__(self, context):
        self.context = context

        # XXX: Explicitly flush collective.indexing queue to make sure we
        # work with up-to-date results.
        processQueue()

        catalog = getToolByName(context, 'portal_catalog')
        self.keywords = catalog.uniqueValuesFor('Subject')
        self.vocab = SimpleVocabulary(
            [SimpleTerm(x, as_keyword_token(x), x) for x in self.keywords])

    def __contains__(self, term):
        return self.vocab.__contains__(term)

    def __iter__(self):
        return self.vocab.__iter__()

    def __len__(self):
        return self.vocab.__len__()

    def getTerm(self, value):
        return self.vocab.getTerm(value)

    def getTermByToken(self, value):
        return self.vocab.getTermByToken(value)

    def search(self, query_string):
        q = query_string.lower()
        return [self.getTerm(kw) for kw in self.keywords if q in kw.lower()]


@implementer(IContextSourceBinder)
class KeywordSearchableSourceBinder(object):

    def __call__(self, context):
        return KeywordSearchableSource(context)


@implementer(IContextSourceBinder)
class KeywordSearchableAndAddableSourceBinder(object):

    def __call__(self, context):
        return KeywordWidgetAddableSourceWrapper(KeywordSearchableSource(context))
